# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for parameters."""

from typing import Generic, NamedTuple, TypeVar

T = TypeVar('T')


class Choice(NamedTuple, Generic[T]):
  """
  A tuple describing a choice parameter.

  Attributes:
    name: A name describing the parameter.
    values: A list of values this parameter may take.
  """

  name: str
  values: list[T]


class LogRange(NamedTuple):
  """
  A tuple describing a logaritmhic range parameter.

  Attributes:
    name: A name describing the parameter.
    bounds: The bounds within the parameter lives.
  """

  name: str
  bounds: tuple[float, float]


class Range(NamedTuple):
  """
  A tuple describing a range parameter.

  Attributes:
    name: A name describing the parameter.
    bounds: The bounds within the parameter lives.
  """

  name: str
  bounds: tuple[float, float]


Parameter = Choice[T] | LogRange | Range



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for mean functions."""

from typing import Protocol

from boax.utils.typing import Array


class Mean(Protocol):
  """
  A callable type for mean functions.

  A mean functions takes a `b0 x ... x bn x n x d`-dim array as input and returns
  an `b0 x ... x bn x n`-dim array of mean function values.
  """

  def __call__(self, index_points: Array) -> Array:
    """
    Calculates the mean function to a set of inputs.

    Args:
      index_points: The inputs to the mean function.

    Returns:
      The function values at the inputs.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for acquisition functions."""

from typing import Generic, Protocol, TypeVar

from boax.utils.typing import Array, PRNGKey

T = TypeVar('T')


class Acquisition(Protocol, Generic[T]):
  """
  A callable type for acquisition functions.

  An acquisition function takes a set of index_points
  as input and returns a numeric acquisition value.
  """

  def __call__(
    self,
    key: PRNGKey,
    model: T,
    observation_index_points: Array,
    observations: Array,
  ) -> tuple[Array, float]:
    """
    Evaluates the acquisition function on a set of index points.

    Args:
      index_points: The set of points the acquisition function is evaluated.

    Returns:
      The acquisition values.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for parameters."""

import math
from collections.abc import Iterator
from functools import partial
from itertools import product
from operator import is_not
from typing import Any

from boax.experiments.search_spaces.base import SearchSpace
from boax.experiments.search_spaces.parameters import (
  Choice,
  LogRange,
  Parameter,
  Range,
  from_dict,
  from_parameterization,
  to_parameterization,
)


def from_dicts(search_space: list[dict]) -> SearchSpace | None:
  """
  Parses a search space from a list of dictionaries each describing a single parameters.

  Args:
    search_space: A list of dictionaries each describing a single parameter.

  Returns:
    The parsed search space with its paramters or None
    if one of the dictionaries could not be parsed.
  """

  choice_parameters = []
  range_parameters = []

  for parameter in search_space:
    match from_dict(parameter):
      case None:
        return None

      case Choice(name, values):
        choice_parameters.append(Choice(name, values))

      case LogRange(name, bounds):
        range_parameters.append(LogRange(name, bounds))

      case Range(name, bounds):
        range_parameters.append(Range(name, bounds))

  return SearchSpace(choice_parameters, range_parameters)


def get_bounds(search_space: SearchSpace) -> Iterator[tuple[float, float]]:
  """
  Extracts the bounds from the search space's range parameters.

  Args:
    search_space: The search space to extract the bounds from.

  Returns:
    An iterator containing the bounds of all range parameters within the search space.
  """

  for parameter in search_space.range_parameters:
    match parameter:
      case LogRange(_, bounds):
        yield math.log(bounds[0]), math.log(bounds[1])

      case Range(_, bounds):
        yield bounds


def get_variants(search_space: SearchSpace) -> Iterator[Any]:
  """
  Extracts all different variants from the search space's choice parameters.

  Args:
    search_space: The search space to extract the bounds from.

  Returns:
    An iterator containing all combinations of the choice parameter values within the search space.
  """

  return product(*(choice.values for choice in search_space.choice_parameters))


def from_parameterizations(
  parameters: list[Parameter], parameterization: dict[str, Any]
) -> list[Any]:
  """
  Parses raw values from a parameterization dictionary.

  Args:
    parameters: A list of parameters of which the raw values should be extracted from.
    parameterization: The dictionary containing paramterization values.

  Returns:
    Parsed raw values.
  """

  return list(
    filter(
      partial(is_not, None),
      map(
        partial(
          from_parameterization,
          parameterization=parameterization,
        ),
        parameters,
      ),
    )
  )


def to_parameterizations(
  parameters: list[Parameter], raw: list[Any]
) -> dict[str, Any]:
  """
  Parses parameterization dictionaries from raw values.

  Args:
    parameter: The parameters to use for parameterizing the raw values.
    raw: The list of raw value to be turned into a paramterization.

  Returns:
    Parsed parameterization dictionary given the parameters and a raw values.
  """

  return dict(
    filter(
      partial(is_not, None),
      map(to_parameterization, parameters, raw),
    )
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Analytic acquisition functions."""

import math

from jax import numpy as jnp
from jax import scipy

from boax.core import distributions
from boax.core.distributions.multivariate_normal import MultivariateNormal
from boax.core.distributions.normal import Normal
from boax.utils.math import log1mexp, logerfcx
from boax.utils.typing import Array, Numeric

inv_sqrt2 = 1 / math.sqrt(2)
c1 = math.log(2 * math.pi) / 2
c2 = math.log(math.pi / 2) / 2


def mean_and_sigma(mvn: MultivariateNormal) -> Normal:
  return distributions.normal.normal(
    mvn.mean, jnp.sqrt(jnp.diagonal(mvn.cov, 0, -2, -1))
  )


def scaled_improvement(normal: Normal, best: Numeric) -> Array:
  return (normal.loc - best) / normal.scale


def ei(normal: Normal, best: Numeric) -> Array:
  u = scaled_improvement(normal, best)
  return (scipy.stats.norm.pdf(u) + u * scipy.stats.norm.cdf(u)) * normal.scale


def lei(normal: Normal, best: Numeric) -> Array:
  def log_ei_upper(x):
    return jnp.log(scipy.stats.norm.pdf(x) + x * scipy.stats.norm.cdf(x))

  def log_ei_lower(x):
    return (
      -(x**2) / 2 - c1 + log1mexp(logerfcx(-x * inv_sqrt2) * jnp.abs(x) + c2)
    )

  def logh(x):
    upper = jnp.where(x > -1.0, x, -1.0)
    lower = jnp.where(x <= -1.0, x, -1.0)

    return jnp.where(x > -1, log_ei_upper(upper), log_ei_lower(lower))

  u = scaled_improvement(normal, best)
  return logh(u) + jnp.log(normal.scale)


def ucb(normal: Normal, beta: Numeric) -> Array:
  return jnp.sqrt(beta) * normal.scale + normal.loc



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for mean functions."""

from jax import numpy as jnp

from boax.core.models.means.base import Mean
from boax.utils.typing import Array, Numeric


def zero() -> Mean:
  """
  The zero mean function.

  Computes `y = f(x) = 0`.

  Example:
    >>> mean = zero()
    >>> ys = mean(xs)

  Returns:
    A zero mean function.
  """

  def mean(x: Array) -> Array:
    return jnp.zeros(x.shape[0:-1])

  return mean


def constant(c: Numeric) -> Mean:
  """
  The constant mean function.

  Computes `y = f(x) = c`.

  Example:
    >>> mean = constant(1.96)
    >>> ys = mean(xs)

  Args:
    c: The constant value.

  Returns:
    A constant mean function.
  """

  def mean(x: Array) -> Array:
    return jnp.full(x.shape[0:-1], c)

  return mean


def linear(scale: Array, bias: Numeric) -> Mean:
  """
  The linear mean function.

  Computes `y = f(x) = scale * x + bias`.

  Example:
    >>> mean = linear(jnp.array([0.2, 3.0]), 5.0)
    >>> ys = mean(xs)

  Args:
    scale: The `d`-dim scale parameter.
    bias: The numeric bias.

  Returns:
    A linear mean function.
  """

  def mean(x: Array) -> Array:
    return jnp.matmul(x, scale) + bias

  return mean



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Optimization functions for experiments."""

from collections import deque
from collections.abc import Callable
from typing import Any, Generic, NamedTuple, TypeVar

from jax import numpy as jnp
from jax import random

from boax.acquisitions import Acquisition
from boax.acquisitions.surrogates import Surrogate
from boax.experiments.base import Step, Trial
from boax.utils.typing import Array, PRNGKey

T = TypeVar('T')
S = TypeVar('S')


class State(NamedTuple, Generic[T]):
  params: T
  observations: list[tuple[dict[str, Any], float]]


def optimize(
  key: PRNGKey,
  warm_up_samples: deque[Array],
  from_parameterization: Callable[[dict[str, Any]], list[Any]],
  to_parameterization: Callable[[list[Any]], dict[str, Any]],
  acquisition: Acquisition[S],
  surrogate: Surrogate[T, S],
) -> Trial[State[T]]:
  next_key, best_key = random.split(key)

  def next(
    step: Step[State[T]] | None = None,
    results: list[tuple[dict[str, Any], float]] = [],
  ) -> tuple[Step[State[T]], list[dict[str, Any]]]:
    timestep, (params, observations) = step or (0, (surrogate.init(), []))
    next_observations = observations + results

    if warm_up_samples:
      next_parameterizations = list(
        map(to_parameterization, warm_up_samples.pop())
      )

      return Step(
        timestep + 1, State(params, next_observations)
      ), next_parameterizations

    else:
      parameterizations, scores = tuple(map(list, zip(*next_observations)))

      observation_index_points = jnp.array(
        list(map(from_parameterization, parameterizations))
      )
      observations = jnp.array(scores)

      next_params = surrogate.update(
        params, observation_index_points, observations
      )
      posterior = surrogate.posterior(
        next_params, observation_index_points, observations
      )

      values, _ = acquisition(
        random.fold_in(next_key, timestep),
        posterior,
        observation_index_points,
        observations,
      )

      next_parameterizations = list(map(to_parameterization, values))

      return Step(
        timestep + 1, State(next_params, next_observations)
      ), next_parameterizations

  def best(step: Step[State[T]]) -> tuple[dict[str, Any], float]:
    parameterizations, scores = tuple(map(list, zip(*step.state.observations)))

    observation_index_points = jnp.array(
      list(map(from_parameterization, parameterizations))
    )
    observations = jnp.array(scores)

    posterior = surrogate.posterior(
      step.state.params, observation_index_points, observations
    )

    values, score = surrogate.best(
      random.fold_in(best_key, step.timestep),
      posterior,
    )

    best_parameterizations = to_parameterization(values)

    return best_parameterizations, score

  return Trial(next, best)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Distances for kernels."""

from jax import numpy as jnp

from boax.utils.typing import Array


def squared(x: Array, y: Array) -> Array:
  """
  Computes the squared distance between vectors `x` and `y`.

  Args:
    x: A vector.
    y: A vector.

  Returns:
    Squared distance.
  """

  adjustment = jnp.mean(x, axis=-2, keepdims=True)

  x_adj = x - adjustment
  x_norm = jnp.sum(jnp.power(x_adj, 2), axis=-1, keepdims=True)
  x_pad = jnp.ones_like(x_norm)

  y_adj = y - adjustment
  y_norm = jnp.sum(jnp.power(y_adj, 2), axis=-1, keepdims=True)
  y_pad = jnp.ones_like(y_norm)

  x_ = jnp.concatenate([-2.0 * x_adj, x_norm, x_pad], axis=-1)
  y_ = jnp.concatenate([y_adj, y_pad, y_norm], axis=-1)

  distance = jnp.matmul(x_, jnp.matrix_transpose(y_))

  return jnp.maximum(distance, 0.0)


def euclidean(x: Array, y: Array) -> Array:
  """
  Computes the euclidean distance between vectors `x` and `y`.

  Args:
    x: A vector.
    y: A vector.

  Returns:
    Euclidean distance.
  """

  distance = jnp.sqrt(squared(x, y))

  return jnp.maximum(distance, 1e-30)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The uniform distribution."""

from typing import NamedTuple

from jax import numpy as jnp
from jax import scipy

from boax.utils.typing import Array


class Uniform(NamedTuple):
  """
  A tuple describing the two parameters of the uniform distribution.

  Attributes:
    a: The minium value parameter.
    b: The maximum value parameter.
  """

  a: Array
  b: Array


def uniform(a: Array = jnp.zeros(()), b: Array = jnp.ones(())) -> Uniform:
  """
  Smart constructor for the uniform distribution.

  Args:
    a: The minium value parameter.
    b: The maximum value parameter.

  Returns:
    The `Uniform` distribution object.
  """

  return Uniform(a, b)


def sample(uniform: Uniform, base_samples: Array) -> Array:
  """
  Samples a uniform distribution based on base samples.

  Args:
    uniform: The uniform distribution.
    base_samples: The uniform distributed base samples.

  Returns:
    The samples from the uniform distribution.
  """

  return uniform.a + (uniform.b - uniform.a) * base_samples


def pdf(uniform: Uniform, values: Array) -> Array:
  """
  Probability density function.

  Args:
    uniform: The uniform distribution.
    values: The values to evaluate.

  Returns:
    The probability density function values.
  """

  return scipy.stats.uniform.pdf(values, uniform.a, uniform.b - uniform.a)


def logpdf(uniform: Uniform, values: Array) -> Array:
  """
  Log probability density function.

  Args:
    uniform: The uniform distribution.
    values: The values to evaluate.

  Returns:
    The log probability density function values.
  """

  return scipy.stats.uniform.logpdf(values, uniform.a, uniform.b - uniform.a)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for evaluator functions."""

from typing import NamedTuple

from jax import jit, lax
from jax import numpy as jnp

from boax.core.distributions.beta import Beta
from boax.policies.believes.base import Belief
from boax.utils.typing import Array


class ActionValues(NamedTuple):
  n: Array
  q: Array


def continuous(num_variants: int) -> Belief[ActionValues, float]:
  """
  The continous belief.

  Stores the number of tries and average reward for each variant.

  Example:
    >>> belief = continous(10)
    >>> params = belief.init()
    >>> updated_params = belief.update(params, variant, reward)
    >>> best_variant = belief.best(updated_params)

  Args:
    num_variants: The number of variants.

  Returns;
    The corresponding `Belief`.
  """

  def init_fn() -> ActionValues:
    return ActionValues(
      jnp.ones(num_variants, dtype=jnp.int32),
      jnp.ones(num_variants, dtype=jnp.float32),
    )

  def update_fn(
    params: ActionValues, variant: int, reward: float
  ) -> ActionValues:
    return ActionValues(
      params.n.at[variant].add(1),
      params.q.at[variant].add(
        (reward - params.q[variant]) / params.n[variant]
      ),
    )

  def best_fn(params: ActionValues) -> tuple[int, float]:
    idx = jnp.argmax(params.n)
    return idx, params.q[idx]

  return Belief(jit(init_fn), jit(update_fn), jit(best_fn))


def binary(num_variants: int) -> Belief[Beta, bool]:
  """
  The binary Beta belief.

  Stores a beta distribution for each variant.

  Example:
    >>> belief = beta(10)
    >>> params = belief.init()
    >>> updated_params = belief.update(params, variant, reward)
    >>> best_variant = belief.best(updated_params)

  Args:
    num_variants: The number of variants.

  Returns;
    The corresponding `Belief`.
  """

  def init_fn() -> Beta:
    return Beta(
      jnp.ones(num_variants, dtype=jnp.int32),
      jnp.ones(num_variants, dtype=jnp.int32),
    )

  def update_fn(params: Beta, variant: int, reward: bool) -> Beta:
    def increment_alpha():
      return Beta(params.a.at[variant].add(1), params.b)

    def increment_beta():
      return Beta(params.a, params.b.at[variant].add(1))

    return lax.cond(
      reward,
      increment_alpha,
      increment_beta,
    )

  def best_fn(params: Beta) -> tuple[int, float]:
    idx = jnp.argmax(params.a / params.b)

    return idx, (params.a / params.b)[idx]

  return Belief(jit(init_fn), jit(update_fn), jit(best_fn))



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for likelihoods."""

from typing import Callable, TypeVar

from jax import jit

from boax.core.objectives.base import Objective
from boax.utils.typing import Array

T = TypeVar('T')


def negative_log_likelihood(
  logpdf_fn: Callable[[T, Array], Array],
) -> Objective[T]:
  """
  The negative log likelihood objective function.

  Example:
    >>> objective = negative_log_likelihood(logpdf_fn)
    >>> nll = objective(prediction, targets)

  Args:
    logpdf_fn: The log probability mass function.

  Returns:
    The negative log likelihood `Objective` function.
  """

  def objective(prediction, targets):
    return -logpdf_fn(prediction, targets)

  return jit(objective)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformation functions for kernels."""

from itertools import repeat
from math import prod
from operator import call

from boax.core.models.kernels.base import Kernel
from boax.utils.typing import Array, Numeric


def scaled(inner: Kernel, amplitude: Numeric) -> Kernel:
  """
  Scales a given kernel.

  Computes `k(x, y) = amplitude * inner(x, y)`.

  Example:
    >>> inner = rbf(jnp.array([0.2, 3.0]))
    >>> kernel = scale(inner, 3.0)
    >>> Kxx = kernel(xs, xs)

  Args:
    kernel: The kernel to be scaled.
    amplitude: The parameter controlling the maximum of the kernel.

  Returns:
    A scaled `Kernel`.
  """

  def kernel(x: Array, y: Array) -> Array:
    return amplitude * inner(x, y)

  return kernel


def additive(*kernels: Kernel) -> Kernel:
  """
  Constructs an additive kernel which sums over a sequence of kernels.

  Computes `k(x, y) = k1(x, y) + k2(x, y) + ... + kn(x, y)`.

  Example:
    >>> kernel = additive(map(rbf, [0.2, 0.3, 0.4]))
    >>> Kxx = kernel(xs, xs)

  Args:
    kernels: The sequence of kernels to sum.

  Returns:
    An additive `Kernel`.
  """

  def kernel(x: Array, y: Array) -> Array:
    return sum(map(call, kernels, repeat(x), repeat(y)))

  return kernel


def product(*kernels: Kernel) -> Kernel:
  """
  Constructs a product kernel which multiplies over a sequence of kernels.

  Computes `k(x, y) = k1(x, y) * k2(x, y) * ... * kn(x, y)`.

  Example:
    >>> kernel = product(map(rbf, [0.2, 0.3, 0.4]))
    >>> Kxx = kernel(xs, xs)

  Args:
    kernels: The sequence of inner kernels.

  Returns:
    A product `Kernel`.
  """

  def kernel(x: Array, y: Array) -> Array:
    return prod(map(call, kernels, repeat(x), repeat(y)))

  return kernel



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformation functions for objective functions."""

from typing import TypeVar

from jax import numpy as jnp

from boax.core.objectives.base import Objective
from boax.utils.typing import Array

T = TypeVar('T')


def penalized(
  objective: Objective[T],
  *terms: Array,
) -> Objective[T]:
  """
  Adds penalty terms to an objective function.

  Args:
    objective: The objective to be penalized.
    terms: The penlization terms added to the objective value.

  Returns:
    A penalized `Objective`.
  """

  def transformed(prediction: T, targets: Array) -> Array:
    return objective(prediction, targets) + jnp.sum(jnp.array(terms))

  return transformed



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for transformation functions."""

from collections.abc import Callable
from functools import reduce
from typing import Any, overload

from boax.core.models.transformations.base import (
  InputTransformation,
  OutcomeTransformation,
)


@overload
def chained(*fns: InputTransformation) -> InputTransformation: ...


@overload
def chained(
  *fns: OutcomeTransformation[Any, Any],
) -> OutcomeTransformation[Any, Any]: ...


def chained(*fns):
  def __reduce_fn(f: Callable, g: Callable) -> Callable:
    def __fn(*args, **kwargs):
      return f(g(*args, **kwargs))

    return __fn

  return reduce(__reduce_fn, reversed(fns))



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for experiment functions."""

from collections import deque
from functools import partial
from typing import Any, TypeVar

from jax import numpy as jnp
from jax import random

from boax import acquisitions, policies
from boax.acquisitions import Acquisition, surrogates
from boax.acquisitions.surrogates import Surrogate
from boax.core import distributions, samplers
from boax.experiments import functions, search_spaces
from boax.experiments.base import Trial
from boax.policies import Policy, believes
from boax.policies.believes import Belief

T = TypeVar('T')
S = TypeVar('S')


def optimization(
  parameters: list[dict[str, Any]],
  *,
  seed: int = 0,
  num_warm_up_steps: int = 12,
  batch_size: int = 1,
  acquisition: Acquisition[S] | None = None,
  surrogate: Surrogate[T, S] | None = None,
) -> Trial[functions.optimization.State[T]]:
  """
  Setup for a bayesian optimization experiment.

  Example:
    >>> experiment = optimization([{'name': 'x', 'type': 'range', 'bounds': [0.0, 1.0]}])

  Args:
    parameters: List of parameters to be optimized via bayesian optimization.
      Each parameter is described by a dictionary with a 'name', a 'type' of either
      fixed, range, or log_range, and a 'value' for fixed parameters,
      or 'bounds' for range and log_range parameters.
    seed: The initial random seed.

  Returns:
    A trail object with next and best functions.

  Raises:
    ValueError: If given parameters cannot be parsed or don't match requirements.
  """

  search_space = search_spaces.from_dicts(parameters)

  if not search_space:
    raise ValueError('One or more of the given parameters is invalid.')

  if not search_space.range_parameters:
    raise ValueError('Found no valid range parameters in given parameter list.')

  if search_space.choice_parameters:
    raise ValueError(
      'Optimization currently does not support choice parameters.'
    )

  bounds = jnp.array(list(search_spaces.get_bounds(search_space)))

  from_parameterization_fn = partial(
    search_spaces.from_parameterizations, search_space.range_parameters
  )

  to_parameterization_fn = partial(
    search_spaces.to_parameterizations, search_space.range_parameters
  )

  sample_key, optimize_key, acquisition_key = random.split(random.key(seed), 3)

  sampler = samplers.halton_uniform(sample_key, (num_warm_up_steps, batch_size))

  samples = sampler(distributions.uniform.uniform(bounds[:, 0], bounds[:, 1]))

  return functions.optimization.optimize(
    key=optimize_key,
    warm_up_samples=deque(samples),
    from_parameterization=from_parameterization_fn,
    to_parameterization=to_parameterization_fn,
    surrogate=surrogate or surrogates.single_task_gaussian_process(bounds),
    acquisition=acquisition
    or acquisitions.q_log_expected_improvement(
      bounds, batch_size, samplers.multivariate_normal(acquisition_key, (256,))
    ),
  )


def bandit(
  parameters: list[dict[str, Any]],
  *,
  seed: int = 0,
  policy: Policy[T] | None = None,
  belief: Belief[T, S] | None = None,
) -> Trial[T]:
  """
  Setup for a multi-arm bandit optimization experiment.

  Example:
    >>> experiment = bandit([{'name': 'arm', 'type': 'choice', 'values': ['left', 'middle', 'right']}])

  Args:
    parameters: List of parameters describing the variants to be optimized.
      Each parameter is described by a dictionary with a 'name', a 'type' of choice,
      and 'values' of the variants for each parameter.
    seed: The initial random seed.
    policy: The policy to be used for optimization.
    belief: The belief to be used for optimization.

  Returns:
    A trail object with next and best functions.

  Raises:
    ValueError: If given parameters cannot be parsed or don't match requirements.
  """

  search_space = search_spaces.from_dicts(parameters)

  if not search_space:
    raise ValueError('One or more of the given parameters is invalid.')

  if not search_space.choice_parameters:
    raise ValueError(
      'Found no valid choice parameters in given parameter list.'
    )

  if search_space.range_parameters:
    raise ValueError('Bandit optimization does not support range parameters.')

  variants = list(search_spaces.get_variants(search_space))

  from_parameterization_fn = partial(
    search_spaces.from_parameterizations, search_space.choice_parameters
  )

  to_parameterization_fn = partial(
    search_spaces.to_parameterizations, search_space.choice_parameters
  )

  return functions.bandit.optimize(
    key=random.key(seed),
    variants=variants,
    from_parameterization=from_parameterization_fn,
    to_parameterization=to_parameterization_fn,
    policy=policy or policies.upper_confidence_bound(2.0),
    belief=belief or believes.continuous(len(variants)),
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The beta distribution."""

from typing import NamedTuple

from jax import numpy as jnp
from jax import scipy

from boax.utils.typing import Array


class Beta(NamedTuple):
  """
  A tuple describing the two shape parameters of the beta distribution.

  Attributes:
    a: The alpha parameter.
    b: The beta parameter.
  """

  a: Array
  b: Array


def beta(a: Array = jnp.ones(()), b: Array = jnp.ones(())) -> Beta:
  """
  Smart constructor for the beta distribution.

  Args:
    a: The alpha parameter.
    b: The beta parameter.

  Returns:
    The `Beta` distribution object.
  """

  return Beta(a, b)


def pdf(beta: Beta, values: Array) -> Array:
  """
  Probability density function.

  Args:
    beta: The beta distribution.
    values: The values to evaluate.

  Returns:
    The probability density function values.
  """

  return scipy.stats.beta.pdf(values, beta.a, beta.b)


def cdf(beta: Beta, values: Array) -> Array:
  """
  Cumulative distribution function.

  Args:
    beta: The beta distribution.
    values: The values to evaluate.

  Returns:
    The cumulative distribution function values.
  """

  return scipy.stats.beta.cdf(values, beta.a, beta.b)


def sf(beta: Beta, values: Array) -> Array:
  """
  Survival function.

  Args:
    beta: The beta distribution.
    values: The values to evaluate.

  Returns:
    The survival function values.
  """

  return scipy.stats.beta.sf(values, beta.a, beta.b)


def logpdf(beta: Beta, values: Array) -> Array:
  """
  Log probability density function.

  Args:
    beta: The beta distribution.
    values: The values to evaluate.

  Returns:
    The log probability density function values.
  """

  return scipy.stats.beta.logpdf(values, beta.a, beta.b)


def logcdf(beta: Beta, values: Array) -> Array:
  """
  Log cumulative distribution function.

  Args:
    beta: The beta distribution.
    values: The values to evaluate.

  Returns:
    The log cumulative distribution function values.
  """

  return scipy.stats.beta.logcdf(values, beta.a, beta.b)


def logsf(beta: Beta, values: Array) -> Array:
  """
  Log survival function.

  Args:
    beta: The beta distribution.
    values: The values to evaluate.

  Returns:
    The log survival function values.
  """

  return scipy.stats.beta.logsf(values, beta.a, beta.b)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for surrogate model functions."""

from typing import Generic, NamedTuple, Protocol, TypeVar

from boax.core.models import Model
from boax.utils.typing import Array, PRNGKey

T = TypeVar('T')
S = TypeVar('S')


class InitFn(Protocol, Generic[T]):
  def __call__(self) -> T:
    pass


class UpdateFn(Protocol, Generic[T]):
  def __call__(
    self, params: T, observation_index_points: Array, observations: Array
  ) -> T:
    pass


class PriorFn(Protocol, Generic[T, S]):
  def __call__(self, params: T) -> Model[S]:
    pass


class PosteriorFn(Protocol, Generic[T, S]):
  def __call__(
    self, params: T, observation_index_points: Array, observations: Array
  ) -> Model[S]:
    pass


class BestFn(Protocol, Generic[S]):
  def __call__(self, key: PRNGKey, model: Model[S]) -> tuple[Array, float]:
    pass


class Surrogate(NamedTuple, Generic[T, S]):
  init: InitFn[T]
  update: UpdateFn[T]
  prior: PriorFn[T, S]
  posterior: PosteriorFn[T, S]
  best: BestFn[S]



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The acquisitions sub-package."""

from . import surrogates as surrogates
from .alias import expected_improvement as expected_improvement
from .alias import log_expected_improvement as log_expected_improvement
from .alias import (
  log_probability_of_improvement as log_probability_of_improvement,
)
from .alias import probability_of_improvement as probability_of_improvement
from .alias import q_expected_improvement as q_expected_improvement
from .alias import q_log_expected_improvement as q_log_expected_improvement
from .alias import q_probability_of_improvement as q_probability_of_improvement
from .alias import q_upper_confidence_bound as q_upper_confidence_bound
from .alias import upper_confidence_bound as upper_confidence_bound
from .base import Acquisition as Acquisition



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for solvers."""

from collections.abc import Callable
from typing import Protocol, Tuple

from boax.utils.typing import Array


class Solver(Protocol):
  """
  A callable type for the solving step of an `Optimizer`.
  """

  def __call__(
    self, fun: Callable[[Array], Array], candidates: Array
  ) -> Tuple[Array, Array]:
    """
    The solving function.

    Args:
      fun: The function to be optimized.
      candidates: The initial guess.

    Returns:
      A tuple of the maxima and their acquisition values.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for acquisition functions."""

from jax import jit, lax, nn, random
from jax import numpy as jnp

from boax.core.distributions.beta import Beta
from boax.policies.base import Policy
from boax.policies.believes import ActionValues
from boax.utils.typing import Numeric, PRNGKey


def epsilon_greedy(epsilon: Numeric) -> Policy[ActionValues]:
  """
  The epsilon greedy policy function.

  Greedily selects the variant with the highest action value with a probability
  of `1 - epsilon` or uniform randomly selects a variant with probability of `epsilon`.

  Example:
    >>> policy = epsilon_greedy(epsilon)
    >>> variant = policy(params, timestep, key)

  Args:
    epsilon: The parameter guiding exploration vs exploitation.

  Returns:
    The corresponding `Policy`.
  """

  def explore(params: ActionValues, key: PRNGKey) -> int:
    return random.choice(key, jnp.arange(len(params.q)))

  def exploit(params: ActionValues, key: PRNGKey) -> int:
    return jnp.argmax(params.q)

  def policy(params: ActionValues, timestep: int, key: PRNGKey) -> int:
    uniform_rng, choice_rng = random.split(key)

    return lax.cond(
      random.uniform(uniform_rng) < epsilon,
      explore,
      exploit,
      params,
      choice_rng,
    )

  return jit(policy)


def boltzmann(tau: Numeric) -> Policy[ActionValues]:
  """
  The boltzmann policy function.

  Randomly selects a variant proportional to the current action-values.

  Example:
    >>> policy = boltzmann(tau)
    >>> variant = policy(params, timestep, key)

  Args:
    tau: The temperature parameter guiding exploration vs exploitation.

  Returns:
    The corresponding `Policy`.
  """

  def policy(params: ActionValues, timestep: int, key: PRNGKey) -> int:
    return random.choice(
      key, jnp.arange(len(params.q)), p=nn.softmax(params.q / tau)
    )

  return jit(policy)


def upper_confidence_bound(confidence: Numeric) -> Policy[ActionValues]:
  """
  The upper confidence bound policy function.

  Selects the variant with highest action-value plus the upper confidence bound.

  Example:
    >>> policy = upper_confidence_bound(confidence)
    >>> variant = policy(params, timestep, key)

  Args:
    confidence: The confidence parameter guiding exploration vs exploitation.

  Returns:
    The corresponding `Policy`.
  """

  def policy(params: ActionValues, timestep: int, key: PRNGKey) -> int:
    return jnp.argmax(
      params.q + confidence * jnp.sqrt(jnp.log(timestep) / params.n)
    )

  return jit(policy)


def thompson_sampling() -> Policy[Beta]:
  """
  The thompson sampling policy function.

  Randomly samples action values for all variants
  and selects the variant with the highest sampled values.

  Example:
    >>> policy = thompson_sampling()
    >>> variant = policy(params, timestep, key)

  Returns:
    The corresponding `Policy`.
  """

  def policy(params: Beta, timestep: int, key: PRNGKey) -> int:
    return jnp.argmax(random.beta(key, params.a, params.b))

  return jit(policy)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The math sub-package."""

import math
from collections.abc import Sequence
from typing import Callable

from jax import lax, nn, scipy
from jax import numpy as jnp

from boax.utils.typing import Array, Numeric

log2 = math.log(2)
sqrt2 = math.sqrt(2)
sqrt3 = math.sqrt(3)

TAU = 1.0
ALPHA = 2.0

Axis = int | Sequence[int] | None


def log1mexp(x: Array) -> Array:
  upper = jnp.where(x > -log2, x, -log2)
  lower = jnp.where(x <= -log2, x, -log2)

  return jnp.where(
    x > -log2, jnp.log(-jnp.expm1(upper)), jnp.log1p(-jnp.exp(lower))
  )


def logerfcx(x: Array) -> Array:
  upper = jnp.where(x > 0.0, x, 0.0)
  lower = jnp.where(x <= 0.0, x, 0.0)

  return jnp.where(
    x > 0,
    jnp.log(jnp.exp(upper**2) * scipy.special.erfc(upper)),
    jnp.log(scipy.special.erfc(lower)) + lower**2,
  )


def logmeanexp(x: Array, *, axis: Axis, keepdims: bool = False) -> Array:
  match axis:
    case None:
      return logsumexp(x, axis=axis, keepdims=keepdims)

    case n if isinstance(n, int):
      return logsumexp(x, axis=axis, keepdims=keepdims) - math.log(x.shape[n])

    case ns:
      return logsumexp(x, axis=axis, keepdims=keepdims) - math.log(
        math.prod(x.shape[n] for n in ns)
      )


def logsumexp(x: Array, *, axis: Axis, keepdims: bool = False) -> Array:
  return _inf_max_helper(
    scipy.special.logsumexp, x=x, axis=axis, keepdims=keepdims
  )


def fatplus(x: Array, alpha: Numeric = 1e-1, tau: Numeric = TAU) -> Array:
  return tau * (nn.softplus(x / tau) + alpha * _cauchy(x / tau))


def fatmax(
  x: Array,
  *,
  axis: Axis,
  keepdims: bool = False,
  alpha: Numeric = ALPHA,
  tau: Numeric = TAU,
) -> Array:
  def max_fun(x: Array, *, axis, keepdims):
    return tau * jnp.log(
      jnp.sum(_pareto(-x / tau, alpha=alpha), axis=axis, keepdims=keepdims)
    )

  return _inf_max_helper(max_fun, x, axis=axis, keepdims=keepdims)


def _inf_max_helper(
  max_fun: Callable[[Array], Array], x: Array, *, axis: Axis, keepdims: bool
) -> Array:
  M = jnp.amax(x, axis=axis, keepdims=True)
  is_inf_max = lax.bitwise_and(M == jnp.inf, M == x)
  has_inf_max = jnp.any(is_inf_max, axis=axis, keepdims=True)

  y_inf = jnp.where(is_inf_max, x, 0.0)
  M_no_inf = jnp.where(M == jnp.inf, 0.0, M)
  y_no_inf = jnp.where(has_inf_max, 0.0, x) - M_no_inf

  result = jnp.where(
    has_inf_max,
    jnp.sum(y_inf, axis=axis, keepdims=True),
    M_no_inf + max_fun(y_no_inf, axis=axis, keepdims=True),
  )

  return result if keepdims else jnp.squeeze(result, axis=axis)


def _pareto(x: Array, *, alpha: Numeric) -> Array:
  beta_1 = alpha
  beta_0 = alpha / 2 * beta_1

  return jnp.power(
    (beta_0 / (beta_0 + beta_1 * x + jnp.square(x))),
    alpha / 2,
  )


def _cauchy(x: Array) -> Array:
  return 1 / (1 + jnp.square(x))



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Outcome transformation functions."""

from collections.abc import Callable
from functools import partial
from typing import TypeVar

from jax import numpy as jnp

from boax.core.models.transformations.base import OutcomeTransformation
from boax.utils.typing import Array

T = TypeVar('T')


def scaled(
  loc: Array, variance: Array, scale_fn: Callable[[T], T]
) -> OutcomeTransformation[T, T]:
  return partial(scale_fn, loc=loc, scale=jnp.sqrt(variance))



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for initializers."""

from collections.abc import Callable
from typing import Protocol

from boax.utils.typing import Array, PRNGKey


class Initializer(Protocol):
  """
  A callable type for the initialization step of an `Optimizer`.
  """

  def __call__(self, fun: Callable[[Array], Array], key: PRNGKey) -> Array:
    """
    The initialization function.

    Args:
      fun: The scoring function.
      key: A PRNG key.

    Returns:
      The initial set of candidates.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for belief functions."""

from typing import Generic, NamedTuple, Protocol, TypeVar

T = TypeVar('T')
R = TypeVar('R')


class InitFn(Protocol, Generic[T]):
  """
  A callable type for policy belief's init functions.

  An init function returns the initial set of parameters.
  """

  def __call__(self) -> T:
    """
    Initialise the parameters.

    Returns:
      The initial parameters.
    """


class UpdateFn(Protocol, Generic[T, R]):
  """
  A callable type for policy belief's update functions.

  An update function takes a set of parameters of type `T`,
  the selected variant, and a reward as input and returns an updated set of parameters.
  """

  def __call__(self, params: T, variant: int, reward: R) -> T:
    """
    Update the parameters.

    Args:
      params: The policy's parameters.
      variant: The selected variant.
      reward: The received reward.

    Returns:
      The updated paramters.
    """


class BestFn(Protocol, Generic[T]):
  """
  A callable type for policy belief's best functions.

  A best function takes a set of parameters of type `T`
  and returns the best variant according to the parameters.
  """

  def __call__(self, params: T) -> tuple[int, float]:
    """
    Update the parameters.

    Args:
      params: The policy's parameters.

    Returns:
      The best variant and its predicted reward.
    """


class Belief(NamedTuple, Generic[T, R]):
  """
  A policy belief.

  The belief is defined by a set of an `init`, `update`, and an `best` function.
  """

  init: InitFn[T]
  update: UpdateFn[T, R]
  best: BestFn[T]



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for samplers."""

from typing import Generic, Protocol, TypeVar

from boax.utils.typing import Array

T = TypeVar('T')


class Sampler(Protocol, Generic[T]):
  """
  A callable type for sampling functions.

  A sampler takes a distribution as its input and samples from it.
  """

  def __call__(self, distribution: T) -> Array:
    """
    Draws samples from the given distribution.

    Args:
      distribution: The distribution to be sampled from.

    Returns:
      A set of samples.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The normal distribution."""

import math
from typing import NamedTuple

from jax import numpy as jnp
from jax import scipy

from boax.utils.typing import Array, Numeric

sqrt2 = math.sqrt(2)


class Normal(NamedTuple):
  """
  A tuple describing the two parameters of the normal distribution.

  Attributes:
    loc: The location parameter.
    scale: The scale parameter.
  """

  loc: Array
  scale: Array


def normal(loc: Array = jnp.zeros(()), scale: Array = jnp.ones(())) -> Normal:
  """
  Smart constructor for the normal distribution.

  Args:
    loc: The location parameter.
    scale: The scale parameter.

  Returns:
    The `Normal` distribution object.
  """

  return Normal(loc, scale)


def sample(normal: Normal, base_samples: Array) -> Array:
  """
  Samples a normal distribution based on base samples.

  Args:
    normal: The normal distribution.
    base_samples: Normal distributed `*sample_shape x *batch_shape x *event_shape`-dim base samples.

  Returns:
    The samples from the normal distribution.
  """

  transposed_base_samples = jnp.transpose(
    base_samples,
    tuple(range(1, base_samples.ndim)) + (0,),
  )

  samples = (
    transposed_base_samples * normal.scale[..., jnp.newaxis]
    + normal.loc[..., jnp.newaxis]
  )

  transposed_samples = jnp.transpose(
    samples, tuple(range(-1, base_samples.ndim - 1))
  )

  return transposed_samples


def scale(normal: Normal, loc: Numeric, scale: Numeric) -> Normal:
  """
  Scales a normal distribution.

  Args:
    normal: The normal distribution.
    loc: The location parameter.
    scale: The scale parameter

  Returns:
    The scaled normal distribution.
  """

  return Normal(normal.loc * scale + loc, normal.scale)


def pdf(normal: Normal, x: Array) -> Array:
  """
  Probability density function.

  Args:
    normal: The normal distribution.
    values: The values to evaluate.

  Returns:
    The probability density function values.
  """

  return scipy.stats.norm.pdf(x, normal.loc, normal.scale)


def cdf(normal: Normal, x: Array) -> Array:
  """
  Cumulative distribution function.

  Args:
    normal: The normal distribution.
    values: The values to evaluate.

  Returns:
    The cumulative distribution function values.
  """
  return scipy.stats.norm.cdf(x, normal.loc, normal.scale)


def sf(normal: Normal, x: Array) -> Array:
  """
  Survival function.

  Args:
    normal: The normal distribution.
    values: The values to evaluate.

  Returns:
    The survival function values.
  """

  return scipy.stats.norm.sf(x, normal.loc, normal.scale)


def logpdf(normal: Normal, x: Array) -> Array:
  """
  Log probability density function.

  Args:
    mvn: The normal distribution.
    values: The values to evaluate.

  Returns:
    The log probability density function values.
  """

  return scipy.stats.norm.logpdf(x, normal.loc, normal.scale)


def logcdf(normal: Normal, x: Array) -> Array:
  """
  Log cumulative distribution function.

  Args:
    normal: The normal distribution.
    values: The values to evaluate.

  Returns:
    The log cumulative distribution function values.
  """

  return scipy.stats.norm.logcdf(x, normal.loc, normal.scale)


def logsf(normal: Normal, x: Array) -> Array:
  """
  Log survival function.

  Args:
    normal: The normal distribution.
    values: The values to evaluate.

  Returns:
    The log survival function values.
  """

  return scipy.stats.norm.logsf(x, normal.loc, normal.scale)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gaussian for surrogate models."""

from typing import TypeVar

from jax import jit
from jax import numpy as jnp

from boax.core.distributions.multivariate_normal import MultivariateNormal
from boax.core.models import functions
from boax.core.models.base import Model
from boax.core.models.kernels.base import Kernel
from boax.core.models.likelihoods.base import Likelihood
from boax.core.models.means.base import Mean
from boax.utils.typing import Array, Numeric

T = TypeVar('T')


def exact(
  mean_fn: Mean,
  kernel_fn: Kernel,
  likelihood_fn: Likelihood[MultivariateNormal, T],
  observation_index_points: Array | None = None,
  observations: Array | None = None,
  jitter: Numeric = 1e-6,
) -> Model[T]:
  """
  The exact gaussian process model.

  Example:
    >>> model = gaussian_process(mean_fn, kernel_fn)
    >>> mean, cov = model(xs)

  Args:
    mean_fn: The process' mean function.
    kernel_fn: The process' covariance function.
    observation_index_points: The index points of the given observations.
    observations: The observed values.
    jitter: The scalar added to the diagonal of the covariance matrix to ensure positive definiteness.

  Returns:
    The gaussian process `Model`.
  """

  if observation_index_points is None or observations is None:

    def model(index_points: Array) -> MultivariateNormal:
      return likelihood_fn(
        functions.gaussian.prior(
          index_points,
          mean_fn,
          kernel_fn,
          jitter,
        )
      )

    return jit(model)

  else:

    def model(index_points: Array) -> MultivariateNormal:
      batch_shape = jnp.broadcast_shapes(
        index_points.shape[:-2],
        observation_index_points.shape[:-2],
      )

      ix = jnp.broadcast_to(index_points, batch_shape + index_points.shape[-2:])

      ox = jnp.broadcast_to(
        observation_index_points,
        batch_shape + observation_index_points.shape[-2:],
      )

      os = jnp.broadcast_to(
        observations,
        batch_shape + observations.shape[-1:],
      )

      return likelihood_fn(
        functions.gaussian.posterior(
          ix,
          ox,
          os,
          mean_fn,
          kernel_fn,
          jitter,
        )
      )

    return jit(model)


# def fantasy(
#   mean_fn: Mean,
#   kernel_fn: Kernel,
#   jitter: Numeric = 1e-6,
# ) -> Callable[[Array, Array, Array], MultivariateNormal]:
#   return vmap(
#     vmap(
#       jit(
#         partial(
#           functions.gaussian.posterior,
#           mean_fn=mean_fn,
#           kernel_fn=kernel_fn,
#           jitter=jitter,
#         )
#       ),
#       in_axes=(0, None, 0),
#     )
#   )


# def multi_fidelity(
#   mean_fn: Mean,
#   kernel_fn: Callable[[Array, Array], Kernel],
#   likelihood_fn: Likelihood[MultivariateNormal, T],
#   observation_index_points: Array | None = None,
#   observation_fidelities: Array | None = None,
#   observations: Array | None = None,
#   jitter: Numeric = 1e-6,
# ) -> Model[T]:
#   """
#   The multi fidelity gaussian process model.

#   Example:
#     >>> model = multi_fidelity(mean_fn, kernel_fn, 1e-4)
#     >>> mean, cov = model(x_train, y_train)(xs)

#   Args:
#     mean_fn: The process' mean function.
#     kernel_fn: The process' covariance function.
#     observation_index_points: The index points of the given observations.
#     observation_fidelities: The fidelities of the given observatons.
#     observations: The observed values.
#     jitter: The scalar added to the diagonal of the covariance matrix to ensure positive definiteness.

#   Returns:
#     The multi fidelity gaussian process `Model`.
#   """

#   if (
#     observation_index_points is None
#     or observation_fidelities is None
#     or observations is None
#   ):
#     return jit(
#       compose(
#         likelihood_fn,
#         partial(
#           functions.multi_fidelity.prior,
#           mean_fn=mean_fn,
#           kernel_fn=kernel_fn,
#           jitter=jitter,
#         ),
#       )
#     )

#   else:
#     return jit(
#       compose(
#         likelihood_fn,
#         partial(
#           functions.multi_fidelity.posterior,
#           observation_index_points=observation_index_points,
#           observation_fidelities=observation_fidelities,
#           observations=observations,
#           mean_fn=mean_fn,
#           kernel_fn=kernel_fn,
#           jitter=jitter,
#         ),
#       )
#     )


# def multi_fidelity_fantasy(
#   mean_fn: Mean,
#   kernel_fn: Callable[[Array, Array], Kernel],
#   jitter: Numeric = 1e-6,
# ) -> Callable[[Array, Array, Array], MultivariateNormal]:
#   fantasy_fn = vmap(
#     vmap(
#       jit(
#         partial(
#           functions.multi_fidelity.posterior,
#           mean_fn=mean_fn,
#           kernel_fn=kernel_fn,
#           jitter=jitter,
#         )
#       ),
#       in_axes=(0, 0, None, None, 0),
#     )
#   )

#   def fn(
#     fantasy_points,
#     observation_index_points,
#     observation_fidelities,
#     observations,
#   ):
#     return fantasy_fn(
#       fantasy_points,
#       jnp.ones_like(fantasy_points),
#       observation_index_points,
#       observation_fidelities,
#       observations,
#     )

#   return fn



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matern kernels."""

import math

from jax import numpy as jnp

from boax.core.models.kernels.functions import distance
from boax.utils.typing import Array, Numeric

sqrt_3 = math.sqrt(3)
sqrt_5 = math.sqrt(5)


def one_half(x: Array, y: Array, length_scale: Numeric) -> Array:
  return jnp.exp(-distance.euclidean(x / length_scale, y / length_scale))


def three_halves(x: Array, y: Array, length_scale: Numeric) -> Array:
  K = sqrt_3 * distance.euclidean(x / length_scale, y / length_scale)
  K = (1.0 + K) * jnp.exp(-K)
  return K


def five_halves(x: Array, y: Array, length_scale: Numeric) -> Array:
  K = sqrt_5 * distance.euclidean(x / length_scale, y / length_scale)
  K = (1.0 + K + K**2 / 3.0) * jnp.exp(-K)
  return K



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for initialization functions."""

from jax import lax, nn, random
from jax import numpy as jnp

from boax.core.optimizers.initializers.base import Initializer
from boax.utils.typing import Array, Numeric


def q_batch(
  samples: Array,
  num_restarts: int,
  eta: Numeric = 1.0,
) -> Initializer:
  """
  Q batch initializer.

  Example:
    >>> initializer = q_batch(fun, samples, num_restarts)
    >>> candidates = initializer(key)

  Args:
    fun: The scoring function.
    samples: The candidate samples.
    num_restarts: The number of restarts.
    eta: The temperature parameter.

  Returns:
    The q-batch `Initializer`.
  """

  def initializer(fun, key):
    y = fun(samples)

    return random.choice(
      key,
      samples,
      (num_restarts,),
      p=jnp.exp(eta * nn.standardize(y, axis=0)),
    )

  return initializer


def q_batch_nonnegative(
  samples: Array,
  num_restarts: int,
  eta: Numeric = 1.0,
  alpha: Numeric = 1e-4,
) -> Initializer:
  """
  Q batch initializer.

  Example:
    >>> initializer = q_batch_nonnegative(fun, samples, num_restarts)
    >>> candidates = initializer(key)

  Args:
    fun: The scoring function.
    samples: The candidate samples.
    num_restarts: The number of restarts.
    eta: The temperature parameter.
    alpha: The alpha parameter.

  Returns:
    The q-batch non-negative `Initializer`.
  """

  def initializer(fun, key):
    y = fun(samples)
    max_val = jnp.max(samples)

    def cond(x):
      return jnp.sum(x[1]) < num_restarts

    def body(x):
      alpha = x[0] * 0.1
      alpha_pos = y >= alpha * max_val

      return (alpha, alpha_pos)

    _, alpha_pos = lax.while_loop(cond, body, (alpha, y >= alpha * max_val))

    return random.choice(
      key,
      samples[alpha_pos],
      (num_restarts,),
      p=jnp.exp(eta * (y[alpha_pos] / max_val - 1)),
    )

  return initializer



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for parameters."""

from typing import Any

from jax import numpy as jnp

from boax.experiments.search_spaces.parameters.alias import (
  Choice,
  LogRange,
  Parameter,
  Range,
)


def from_dict(parameter: dict) -> Parameter | None:
  """
  Parses a parameter from a dictionary.

  Example:
    >>> from_dict({'name': 'colours', 'type': 'choice', 'values': ['red', 'green', 'blue']})
    >>> from_dict({'name': 'learning_rate', 'type': 'log_range', 'bounds': [1e-4, 1e-3]})
    >>> from_dict({'name': 'click_rate', 'type': 'range', 'bounds': [0, 1]})

  Args:
    parameter: A dictionary describing the parameter.

  Returns:
    The parsed parameter or None if parameter could not be parsed.
  """

  match parameter.get('type'):
    case None:
      return None

    case 'choice' if 'name' in parameter and 'values' in parameter:
      return Choice(parameter['name'], parameter['values'])

    case 'log_range' if 'name' in parameter and 'bounds' in parameter:
      return LogRange(parameter['name'], parameter['bounds'])

    case 'range' if 'name' in parameter and 'bounds' in parameter:
      return Range(parameter['name'], parameter['bounds'])


def from_parameterization(
  parameter: Parameter, parameterization: dict[str, Any]
) -> Any | None:
  """
  Parses raw values from a parameterization dictionary.

  Args:
    parameter: The parameter of which the raw value should be extracted from.
    parameterization: The dictionary containing paramterization values.

  Returns:
    Parsed raw value or None if the parameter could not be found in the parameterization.
  """

  match parameter:
    case Choice(name, values) if (
      name in parameterization and parameterization[name] in values
    ):
      return parameterization[name]

    case LogRange(name, bounds) if name in parameterization:
      return jnp.log(jnp.clip(parameterization[name], bounds[0], bounds[1]))

    case Range(name, bounds) if name in parameterization:
      return jnp.clip(parameterization[name], bounds[0], bounds[1])

    case _:
      return None


def to_parameterization(
  parameter: Parameter, raw: Any
) -> tuple[str, Any] | None:
  """
  Parses parameterization dictionaries from raw values.

  Args:
    parameter: The parameter to use for parameterizing the raw value.
    raw: The raw value to be turned into a paramterization.

  Returns:
    Parsed parameterization dictionary given a parameter and a raw value.
  """

  match parameter:
    case Choice(name, values) if raw in values:
      return name, raw

    case LogRange(name, bounds):
      return name, float(jnp.clip(jnp.exp(raw), bounds[0], bounds[1]))

    case Range(name, bounds):
      return name, float(jnp.clip(raw, bounds[0], bounds[1]))

    case _:
      return None



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for solver functions."""

from jax import numpy as jnp
from jax.scipy import optimize

from boax.core.optimizers.solvers.base import Solver
from boax.utils.typing import Array


def scipy(
  bounds: Array,
  method: str = 'bfgs',
) -> Solver:
  """
  Scipy solver.

  Example:
    >>> solver = scipy(fun, bounds)
    >>> next_candidates, values = solver(candidates)

  Args:
    bounds: The bounds of the search space.
    method: The solver method.

  Returns:
    The scipy `Solver`.
  """

  def solver(fun, candidates):
    def score(x: Array) -> Array:
      return jnp.negative(jnp.sum(fun(jnp.reshape(x, candidates.shape))))

    results = optimize.minimize(
      fun=score,
      x0=candidates.flatten(),
      method=method,
    )

    clipped = jnp.clip(
      jnp.reshape(results.x, candidates.shape),
      min=bounds[:, 0],
      max=bounds[:, 1],
    )

    return clipped, fun(clipped)

  return solver



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bandit functions for experiments."""

from collections.abc import Callable
from typing import Any, TypeVar

from jax import random

from boax.experiments.base import Step, Trial
from boax.policies import Policy
from boax.policies.believes import Belief
from boax.utils.typing import PRNGKey

T = TypeVar('T')
S = TypeVar('S')


def optimize(
  key: PRNGKey,
  variants: list[Any],
  policy: Policy[T],
  belief: Belief[T, S],
  from_parameterization: Callable[[dict[str, Any]], list[Any]],
  to_parameterization: Callable[[list[Any]], dict[str, Any]],
) -> Trial[T]:
  def next(
    step: Step[T] | None = None,
    results: list[tuple[dict[str, Any], float]] = [],
  ) -> tuple[Step[T], list[dict[str, Any]]]:
    timestep, state = step or (0, belief.init())

    for parameterization, score in results:
      variant = tuple(from_parameterization(parameterization))

      if variant in variants:
        state = belief.update(
          state,
          variants.index(variant),
          score,
        )

    idx = policy(
      state,
      timestep,
      random.fold_in(key, timestep),
    )
    parameterizations = [to_parameterization(variants[int(idx)])]

    return Step(timestep + 1, state), parameterizations

  def best(step: Step[T]) -> tuple[dict[str, Any], float]:
    idx, reward = belief.best(step.state)
    parameterizations = to_parameterization(variants[int(idx)])

    return parameterizations, reward

  return Trial(next, best)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for surrogate models."""

from typing import Generic, Protocol, TypeVar

from boax.utils.typing import Array

T = TypeVar('T')


class Model(Protocol, Generic[T]):
  """
  A callable type for surrogate models.

  A model takes a set of `b0 x ... x bn x n x d`-dim index points as input
  and returns a posterior prediction of type `T`.
  """

  def __call__(self, index_points: Array) -> T:
    """
    Computes the posterior prediction at the index points.

    Args:
      index_points: The `b0 x ... x bn x n x d` index points.

    Returns:
      The model evaluated at the given index points.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monte Carlo acquisition functions."""

from jax import lax, nn
from jax import numpy as jnp

from boax.utils.math import fatplus
from boax.utils.typing import Array, Numeric


def qei(values: Array, best: Numeric) -> Array:
  return lax.max(values - best, 0.0)


def qlei(values: Array, best: Numeric, tau: Numeric) -> Array:
  return jnp.log(fatplus(values - best, tau=tau))


def qpoi(values: Array, best: Numeric, tau: Numeric) -> Array:
  return nn.sigmoid((values - best) / tau)


def qucb(values: Array, beta: Numeric) -> Array:
  mean = jnp.mean(values)
  return mean + beta * jnp.abs(values - mean)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for kernels."""

from functools import partial

from boax.core.models.kernels import functions
from boax.core.models.kernels.base import Kernel
from boax.utils.typing import Array


def rbf(length_scale: Array) -> Kernel:
  """
  The Radial basis function (RBF) kernel.

  Computes `k(x, y) = exp(-||x - y||**2 / (2 * length_scale**2))`.

  Example:
    >>> kernel = rbf(jnp.array([0.2, 3.0]))
    >>> Kxx = kernel(xs, xs)

  Args:
    length_scale: The parameter controlling how sharp or wide the kernel is.

  Returns:
    A RBF `Kernel`.
  """

  return partial(
    functions.rbf.rbf,
    length_scale=length_scale,
  )


def matern_one_half(length_scale: Array) -> Kernel:
  """
  The Matern kernel with parameter 1/2.

  Computes `k(x, y) = exp(-||x - y|| / length_scale)`.

  Example:
    >>> kernel = matern_one_half(jnp.array([0.2, 3.0]))
    >>> Kxx = kernel(xs, xs)

  Args:
    length_scale: The parameter controlling how sharp or wide the kernel is.

  Returns:
    A matern one half `Kernel`.
  """

  return partial(
    functions.matern.one_half,
    length_scale=length_scale,
  )


def matern_three_halves(length_scale: Array) -> Kernel:
  """
  The Matern kernel with parameter 3/2.

  Computes `k(x, y) = (1 + z) * exp(-z)`,

  with `z = sqrt(3) * ||x - y|| / length_scale`.

  Example:
    >>> kernel = matern_three_halves(jnp.array([0.2, 3.0]))
    >>> Kxx = kernel(xs, xs)

  Args:
    length_scale: The parameter controlling how sharp or wide the kernel is.

  Returns:
    A matern three halves `Kernel`.
  """

  return partial(
    functions.matern.three_halves,
    length_scale=length_scale,
  )


def matern_five_halves(length_scale: Array) -> Kernel:
  """
  The Matern kernel with parameter 5/2.

  Computes `k(x, y) = (1 + z + z**2 / 3) * exp(-z)`,

  with `z = sqrt(5) * ||x - y|| / length_scale`.

  Example:
    >>> kernel = matern_five_halves(jnp.array([0.2, 3.0]))
    >>> Kxx = kernel(xs, xs)

  Args:
    length_scale: The parameter controlling how sharp or wide the kernel is.

  Returns:
    A matern five halves `Kernel`.
  """

  return partial(
    functions.matern.five_halves,
    length_scale=length_scale,
  )



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for evaluator functions."""

from operator import attrgetter

import optax
from jax import jit, lax, nn, random, value_and_grad
from jax import numpy as jnp

from boax.acquisitions.surrogates.base import Surrogate
from boax.core import distributions, models, objectives, optimizers, samplers
from boax.core.distributions import multivariate_normal, normal
from boax.core.distributions.multivariate_normal import MultivariateNormal
from boax.core.models import Model
from boax.core.objectives import Objective
from boax.utils.math import sqrt2, sqrt3
from boax.utils.typing import Array, PRNGKey


def single_task_gaussian_process(
  bounds: Array,
  fit_learning_rate: float = 0.01,
  fit_num_steps: int = 500,
  optimizer_num_raw_samples: int = 512,
  optimizer_num_restarts: int = 10,
) -> Surrogate[dict, MultivariateNormal]:
  """
  The single task gaussian process surrogate model.

  Has strict priors on its paramters.

  Example:
    >>> surrogate = single_task_gaussian_process(bounds, learning_rate, num_steps)
    >>> params = surrogate.init()
    >>> fitted_params = surrogate.update(params, observation_index_points, observations)
    >>> prior = surrogate.prior(params)
    >>> posterior = surrogate.update(params, observation_index_points, observations)
    >>> best = surrogate.best(posterior)

  Args:
    bounds: The domain bounds.
    learning_rate: The learning for fitting the parameters during parameter update.
    num_steps: The number of steps for fitting the parameters during parameter update.

  Returns;
    The corresponding `Surrogate`.
  """

  optimizer = optax.adam(fit_learning_rate)

  priors = {
    'length_scale': normal.normal(
      sqrt2 + jnp.log(bounds.shape[0]) * 0.5,
      sqrt3,
    ),
    'noise': normal.normal(
      jnp.array(-4.0),
      jnp.array(1.0),
    ),
  }

  def _model_fn(
    params: dict,
    observation_index_points: Array | None,
    observations: Array | None,
  ) -> Model[MultivariateNormal]:
    return models.gaussian_process.exact(
      models.means.constant(params['mean']),
      models.kernels.rbf(jnp.exp(params['length_scale'])),
      models.likelihoods.gaussian(jnp.exp(params['noise'])),
      observation_index_points,
      observations,
    )

  def _objective_fn(params: dict) -> Objective[MultivariateNormal]:
    return objectives.transformations.penalized(
      objectives.negative_log_likelihood(multivariate_normal.logpdf),
      -jnp.sum(
        normal.logpdf(priors['length_scale'], jnp.exp(params['length_scale'])),
      ),
      -jnp.sum(
        normal.logpdf(priors['noise'], jnp.exp(params['noise'])),
      ),
    )

  def _loss_fn(
    params: dict, observation_index_points: Array, observations: Array
  ) -> Array:
    y_hat = _model_fn(params, None, None)(observation_index_points)
    objective = _objective_fn(params)
    return objective(y_hat, observations)

  def init() -> dict:
    return {
      'mean': jnp.zeros(()),
      'length_scale': jnp.repeat(priors['length_scale'].loc, bounds.shape[0]),
      'noise': priors['noise'].loc,
    }

  def update(
    params: dict, observation_index_points: Array, observations: Array
  ) -> dict:
    def fit(state, i):
      loss, grads = value_and_grad(_loss_fn)(
        state[0], observation_index_points, observations
      )
      updates, opt_state = optimizer.update(grads, state[1])
      params = optax.apply_updates(state[0], updates)

      return (params, opt_state), loss

    (fitted_params, _), _ = lax.scan(
      jit(fit),
      (params, optimizer.init(params)),
      jnp.arange(fit_num_steps),
    )

    return fitted_params

  def prior(params: dict) -> Model[MultivariateNormal]:
    return models.transformations.transformed(
      _model_fn(params, None, None),
      input_transformation_fn=models.transformations.input.normalized(bounds),
    )

  def posterior(
    params: dict, observation_index_points: Array, observations: Array
  ) -> Model[MultivariateNormal]:
    index_points_normalized = (observation_index_points - bounds[:, 0]) / (
      bounds[:, 1] - bounds[:, 0]
    )
    observations_mean, observations_var = (
      jnp.mean(observations),
      jnp.var(observations),
    )
    observations_standardized = nn.standardize(
      observations, mean=observations_mean, variance=observations_var
    )

    return models.transformations.transformed(
      _model_fn(params, index_points_normalized, observations_standardized),
      input_transformation_fn=models.transformations.input.normalized(bounds),
      outcome_transformation_fn=models.transformations.outcome.scaled(
        observations_mean, observations_var, multivariate_normal.scale
      ),
    )

  def best(
    key: PRNGKey, model: Model[MultivariateNormal]
  ) -> tuple[Array, float]:
    sampler_key, optimizer_key = random.split(key)

    samples = samplers.halton_uniform(
      sampler_key, (optimizer_num_raw_samples,)
    )(
      distributions.uniform.uniform(bounds[:, 0], bounds[:, 1]),
    )

    fun = models.transformations.transformed(
      model,
      outcome_transformation_fn=attrgetter('mean'),
    )

    optimizer = optimizers.batch(
      optimizers.initializers.q_batch(samples, optimizer_num_restarts),
      optimizers.solvers.scipy(bounds),
    )

    next_candidate, value = optimizer(fun, optimizer_key)

    return next_candidate, float(value)

  return Surrogate(init, update, prior, posterior, best)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for benchmarks."""

from typing import Callable, NamedTuple

from boax.utils.typing import Array, Numeric


class Benchmark(NamedTuple):
  """
  A tuple describing a benchmark function for bayesian optimization.

  Attributes:
    name: The name of the benchmark function.
    bounds: The bounds within the function is optimized.
    optima: The parameterizations that achive the optimum.
    objective: The benchmark function.
  """

  name: str
  bounds: Array
  optima: Array | list[Array]
  objective: Callable[[Array], Numeric]



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for likelihoods."""

from jax import nn

from boax.core import distributions
from boax.core.distributions.beta import Beta
from boax.utils.typing import Array


def beta(samples: Array, scale: Array) -> Beta:
  mixture = nn.sigmoid(samples)
  alpha = mixture * scale + 1
  beta = scale - alpha + 2

  return distributions.beta.beta(alpha, beta)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for optimizers."""

from jax import numpy as jnp
from jax import random

from boax.core.optimizers.base import Optimizer
from boax.core.optimizers.initializers.base import Initializer
from boax.core.optimizers.solvers.base import Solver


def batch(initializer: Initializer, solver: Solver) -> Optimizer:
  """
  Batch optimizer.

  Example:
    >>> optimizer = batch(initializer, solver)
    >>> next_candidates = optimizer(key)

  Args:
    initializer: The initializer function.
    solver: The solver function.

  Returns:
    The batch `Optimizer`.
  """

  def optimizer(fun, key):
    candidates = initializer(fun, key)
    next_candidates, values = solver(fun, candidates)
    idx = jnp.argmax(values)

    return next_candidates[idx], values[idx]

  return optimizer


def sequential(initializer: Initializer, solver: Solver, q: int) -> Optimizer:
  """
  Sequential optimizer.

  Example:
    >>> optimizer = sequential(initializer, solver)
    >>> next_candidates = optimizer(key, fun, bounds, q, num_samples, num_restarts)

  Args:
    initializer: The initializer function.
    solver: The solver function.

  Returns:
    The sequential `Optimizer`.
  """

  inner = batch(initializer, solver)

  def optimizer(fun, key):
    next_candidates, values = zip(
      *(inner(fun, random.fold_in(key, i)) for i in range(q))
    )

    return (jnp.concatenate(list(next_candidates)), jnp.array(list(values)))

  return optimizer



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The poisson distribution."""

from typing import NamedTuple

from jax import scipy

from boax.utils.typing import Array


class Poisson(NamedTuple):
  """
  A tuple describing the rate parameter of the poisson distribution.

  Attributes:
    mu: The rate parameter.
  """

  mu: Array


def poisson(mu: Array) -> Poisson:
  """
  Smart constructor for the poisson distribution.

  Args:
    mu: The rate parameter.

  Returns:
    The `Poisson` distribution object.
  """

  return Poisson(mu)


def pmf(poisson: Poisson, values: Array) -> Array:
  """
  Probability mass function.

  Args:
    poisson: The poisson distribution.
    values: The values to evaluate.

  Returns:
    The probability mass function values.
  """

  return scipy.stats.poisson.pmf(values, poisson.mu)


def cdf(poisson: Poisson, values: Array) -> Array:
  """
  Cumulative distribution function.

  Args:
    poisson: The poisson distribution.
    values: The values to evaluate.

  Returns:
    The cumulative distribution function values.
  """

  return scipy.stats.poisson.cdf(values, poisson.mu)


def logpmf(poisson: Poisson, values: Array) -> Array:
  """
  Log probability density function.

  Args:
    poisson: The poisson distribution.
    values: The values to evaluate.

  Returns:
    The log probability density function values.
  """

  return scipy.stats.poisson.logpdf(values, poisson.mu)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for sampling functions."""

import numpy as np
from jax import numpy as jnp
from jax import scipy

from boax.utils.math import sqrt2
from boax.utils.typing import Array


def ratio_of_uniforms(base_samples: Array):
  v = 0.5 + (1 - jnp.finfo(base_samples.dtype).eps) * (base_samples - 0.5)
  return scipy.special.erfinv(2 * v - 1) * sqrt2


def primes_less_than(n) -> np.ndarray:
  """
  Sorted array of primes such that `2 <= prime < n`.

  Args:
    n: The upper bound for primes.

  Returns:
    The sorted array of primes.
  """

  j = 3
  primes = np.ones((n + 1) // 2, dtype=bool)

  while j * j <= n:
    if primes[j // 2]:
      primes[j * j // 2 :: j] = False
    j += 2

  ret = 2 * np.where(primes)[0] + 1
  ret[0] = 2  # :(

  return ret



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for objectives."""

from typing import Generic, Protocol, TypeVar

from boax.utils.typing import Array

T = TypeVar('T')


class Objective(Protocol, Generic[T]):
  """
  A callable type for objectives.

  An objective function takes a posterior prediction of type `T`
  and an array of targets as input and returns an objective value.
  """

  def __call__(self, prediction: T, targets: Array) -> Array:
    """
    Computes the objective value at the given prediction of type `T`.

    Args:
      prediction: The prediction.
      targets: The targets.

    Returns:
      The objective values.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The multivariate normal distribution."""

from typing import NamedTuple

from jax import numpy as jnp
from jax import scipy

from boax.utils.typing import Array, Numeric


class MultivariateNormal(NamedTuple):
  """
  A tuple describing the two parameters of the multivariate normal distribution.

  Attributes:
    mean: The mean vector parameter.
    cov: The covariance matrix parameter.
  """

  mean: Array
  cov: Array


def multivariate_normal(
  mean: Array = jnp.zeros((1,)), cov: Array = jnp.identity(1)
) -> MultivariateNormal:
  """
  Smart constructor for the multivariate normal distribution.

  Args:
    mean: The mean vector parameter.
    cov: The covariance matrix parameter.

  Returns:
    The `MultivariateNormal` distribution object.
  """

  return MultivariateNormal(mean, cov)


def sample(mvn: MultivariateNormal, base_samples: Array) -> Array:
  """
  Samples a multivariate normal distribution based on base samples.

  Args:
    mvn: The multivariate normal distribution.
    base_samples: Normal distributed `*sample_shape x *batch_shape x *event_shape`-dim base samples.

  Returns:
    The samples from the multivariate normal distribution.
  """

  cov_root = scipy.linalg.cholesky(mvn.cov, lower=True)

  transposed_base_samples = jnp.transpose(
    base_samples,
    tuple(range(1, base_samples.ndim)) + (0,),
  )

  samples = (
    jnp.matmul(cov_root, transposed_base_samples) + mvn.mean[..., jnp.newaxis]
  )

  transposed_samples = jnp.transpose(
    samples, tuple(range(-1, base_samples.ndim - 1))
  )

  return transposed_samples


def scale(
  mvn: MultivariateNormal, loc: Numeric, scale: Numeric
) -> MultivariateNormal:
  """
  Scales a normal distribution.

  Args:
    normal: The normal distribution.
    loc: The location parameter.
    scale: The scale parameter

  Returns:
    The scaled normal distribution.
  """

  return MultivariateNormal(mvn.mean * scale + loc, mvn.cov)


def pdf(mvn: MultivariateNormal, values: Array) -> Array:
  """
  Probability density function.

  Args:
    mvn: The multivariate normal distribution.
    values: The values to evaluate.

  Returns:
    The probability density function values.
  """

  return scipy.stats.multivariate_normal.pdf(values, mvn.mean, mvn.cov)


def logpdf(mvn: MultivariateNormal, values: Array) -> Array:
  """
  Log probability density function.

  Args:
    mvn: The multivariate normal distribution.
    values: The values to evaluate.

  Returns:
    The log probability density function values.
  """

  return scipy.stats.multivariate_normal.logpdf(values, mvn.mean, mvn.cov)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multi fidelity functions."""

from typing import Callable

from jax import numpy as jnp
from jax import scipy

from boax.core import distributions
from boax.core.distributions.multivariate_normal import MultivariateNormal
from boax.core.models.kernels.base import Kernel
from boax.core.models.means.base import Mean
from boax.utils.typing import Array, Numeric


def prior(
  index_points: Array,
  fidelities: Array,
  mean_fn: Mean,
  kernel_fn: Callable[[Array, Array], Kernel],
  jitter: Numeric,
) -> MultivariateNormal:
  Kxx = kernel_fn(fidelities, fidelities)(index_points, index_points)
  mean = mean_fn(index_points)
  cov = Kxx + jitter * jnp.identity(Kxx.shape[-1])

  return distributions.multivariate_normal.multivariate_normal(mean, cov)


def posterior(
  index_points: Array,
  fidelities: Array,
  observation_index_points: Array,
  observation_fidelities: Array,
  observations: Array,
  mean_fn: Mean,
  kernel_fn: Callable[[Array, Array], Kernel],
  jitter: Numeric,
) -> MultivariateNormal:
  mz = mean_fn(index_points)
  mx = mean_fn(observation_index_points)

  Kxx = kernel_fn(observation_fidelities, observation_fidelities)(
    observation_index_points, observation_index_points
  )

  Kxz = kernel_fn(observation_fidelities, fidelities)(
    observation_index_points, index_points
  )

  Kzz = kernel_fn(fidelities, fidelities)(index_points, index_points)

  K = Kxx + jitter * jnp.identity(Kxx.shape[-1])
  chol = scipy.linalg.cholesky(K, lower=True)
  kinvy = scipy.linalg.solve_triangular(
    chol.T, scipy.linalg.solve_triangular(chol, observations - mx, lower=True)
  )
  v = scipy.linalg.solve_triangular(chol, Kxz, lower=True)

  mean = mz + jnp.dot(Kxz.T, kinvy)
  cov = Kzz - jnp.dot(v.T, v) + jitter * jnp.identity(Kzz.shape[-1])

  return distributions.multivariate_normal.multivariate_normal(mean, cov)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformation functions for models."""

from itertools import repeat
from operator import call
from typing import Any, Sequence, TypeVar

from boax.core.models.base import Model
from boax.core.models.transformations.base import (
  InputTransformation,
  OutcomeTransformation,
)
from boax.utils.typing import Array

A = TypeVar('A')
B = TypeVar('B')
T = TypeVar('T')


def transformed(
  model: Model[A],
  *,
  input_transformation_fn: InputTransformation | None = None,
  outcome_transformation_fn: OutcomeTransformation[A, B] | None = None,
) -> Model[B]:
  """
  Constructs an model with input_transformation_fn applied to its inputs and
  the output_transformation_fn applied to its outcomes.

  Example:
    >>> transformed = transformed(
      model,
      input_transformation_fn=fn1,
      outcome_transformation_fn=fn2
    )
    >>> result = transformed(xs)

  Args:
    model: The base model.
    input_transformation_fn: The transformation functions applied to the model's inputs.
    outcome_transformation_fn: The transformation functions applied to the model's outcomes.

  Returns:
    The transformed `Model` function.
  """

  match (input_transformation_fn, outcome_transformation_fn):
    case (None, None):
      return model

    case (fn, None):

      def input_transformed(x: Array) -> Array:
        return model(fn(x))

      return input_transformed

    case (None, fn):

      def output_transformed(x: Array) -> Array:
        return fn(model(x))

      return output_transformed

    case (fn1, fn2):

      def input_output_transformed(x: Array) -> Array:
        return fn2(model(fn1(x)))

      return input_output_transformed


def joined(*models: Model[Any]) -> Model[Sequence[Any]]:
  """
  Constructs a joined model.

  Example:
    >>> transformed = joined(objective_model, cost_model)
    >>> objective_result, cost_result = transformed(xs)

  Args:
    models: The models to be joined.

  Returns:
    The transformed `Model` function.
  """

  def model(x: Array) -> Array:
    tuple(map(call, models, repeat(x)))

  return model



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for benchmark functions."""

from jax import jit
from jax import numpy as jnp

from boax.experiments.benchmarks import functions
from boax.experiments.benchmarks.base import Benchmark

beale = Benchmark(
  name='Beale',
  bounds=jnp.array([[-4.5, 4.5]] * 2),
  optima=jnp.array([3.0, 0.5]),
  objective=jit(functions.beale),
)

bohachevsky = Benchmark(
  name='Bohachevsky',
  bounds=jnp.array([[-100.0, 100.0]] * 2),
  optima=jnp.array([0, 0]),
  objective=jit(functions.bohachevsky),
)

branin = Benchmark(
  name='Branin',
  bounds=jnp.array([[-5.0, 10.0], [0.0, 15.0]]),
  optima=list[
    jnp.array([-jnp.pi, 12.275]),
    jnp.array([jnp.pi, 2.275]),
    jnp.array([9.42478, 2.475]),
  ],
  objective=jit(functions.branin),
)

forrester_1d = Benchmark(
  name='Forrester',
  bounds=jnp.array([[0.0, 1.0]]),
  optima=jnp.array([0.757249]),
  objective=jit(functions.forrester_1d),
)

grammacy_lee = Benchmark(
  name='Grammacy & Lee',
  bounds=jnp.array([[0.5, 2.5]]),
  optima=jnp.array([-0.86901113]),
  objective=jit(functions.grammacy_lee),
)

hartmann_6d = Benchmark(
  name='Hartmann 6D',
  bounds=jnp.array([[0.0, 1.0]] * 6),
  optima=jnp.array([0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573]),
  objective=jit(functions.hartmann_6d),
)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quasi Random sampling functions."""

from typing import Sequence

from jax import numpy as jnp
from jax import random

from boax.core.samplers.functions.utils import primes_less_than
from boax.utils.typing import Array, PRNGKey

# The maximum dimension we support. This is limited by the number of primes in the PRIMES array.
MAX_DIMENSION = 10000
PRIMES = primes_less_than(104729 + 1)
assert len(PRIMES) == MAX_DIMENSION


def halton_sequence(
  key: PRNGKey, sample_shape: Sequence[int], ndims: int
) -> Array:
  shuffle_key, correction_key = random.split(key)
  num_samples = jnp.prod(jnp.asarray(sample_shape))

  radixes = PRIMES[0:ndims][..., jnp.newaxis]
  indices = jnp.reshape(jnp.arange(num_samples) + 1, (-1, 1, 1))

  max_sizes_by_axes = jnp.floor(jnp.log(num_samples) / jnp.log(radixes) + 1)
  max_size = jnp.max(max_sizes_by_axes)

  exponents_by_axes = jnp.tile(
    jnp.array([jnp.arange(max_size)]), jnp.array([ndims, 1])
  )
  weight_mask = exponents_by_axes < max_sizes_by_axes
  capped_exponents = jnp.where(weight_mask, exponents_by_axes, jnp.zeros(()))
  weights = radixes**capped_exponents

  coeffs = jnp.floor_divide(indices, weights) * weight_mask % radixes
  shuffled = halton_shuffle(shuffle_key, coeffs, radixes) * weight_mask

  base_values = jnp.sum(shuffled / (radixes * weights), axis=-1)
  zero_correction = random.uniform(correction_key, (ndims, 1))

  return jnp.reshape(
    base_values + (zero_correction / (radixes**max_sizes_by_axes)).flatten(),
    sample_shape + (ndims,),
  )


def halton_shuffle(key: PRNGKey, coeffs: Array, radixes: Array) -> Array:
  icoeffs = jnp.astype(coeffs, jnp.int32)
  iradixes = jnp.astype(radixes, jnp.int32)
  num_coeffs = coeffs.shape[-1]

  permutations = halton_permutations(key, radixes.flatten(), num_coeffs)
  radix_offsets = jnp.reshape(
    jnp.hstack([jnp.array(0), jnp.cumsum(iradixes[:-1])]), (-1, 1)
  )
  offsets = radix_offsets + jnp.arange(num_coeffs) * jnp.sum(iradixes)

  return permutations[icoeffs + offsets]


def halton_permutations(key: PRNGKey, dims: Array, num_results: int) -> Array:
  max_size = jnp.max(dims)
  mask = jnp.arange(max_size) >= dims[..., jnp.newaxis]

  indices = jnp.stack(jnp.where(~jnp.tile(mask, (num_results, 1, 1))), axis=-1)

  samples = jnp.argsort(
    jnp.where(
      mask,
      jnp.arange(max_size) + 10.0,
      random.uniform(key, (num_results, dims.size, max_size)),
    )
  )

  return samples[*indices.T]



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The benchmark function implementations."""

from jax import numpy as jnp

from boax.utils.typing import Array, Numeric


def beale(x: Array) -> Numeric:
  a = 1.5 - x[..., 0] + x[..., 0] * x[..., 1]
  b = 2.25 - x[..., 0] + x[..., 0] * (x[..., 1] ** 2)
  c = 2.625 - x[..., 0] + x[..., 0] * (x[..., 1] ** 3)

  return (a**2) + (b**2) + (c**2)


def bohachevsky(x: Array) -> Numeric:
  a = x[..., 0] ** 2
  b = 2 * x[..., 1] ** 2
  c = 0.3 * jnp.cos(3 * jnp.pi * x[..., 0])
  d = 0.4 * jnp.cos(4 * jnp.pi * x[..., 1])

  return a + b - c - d + 0.7


def branin(x: Array) -> Numeric:
  a = (
    x[..., 1]
    - 5.1 / 4 * jnp.pi**2 * x[..., 0] ** 2
    + 5 / jnp.pi * x[..., 0]
    - 6
  )
  b = 10 * (1 - 1 / 8 * jnp.pi) * jnp.cos(x[..., 0])

  return a**2 + b + 10


def forrester_1d(x: Array) -> Numeric:
  return (6 * x[..., 0] - 2) ** 2 * jnp.sin(12 * x[..., 0] - 4)


def grammacy_lee(x: Array) -> Numeric:
  return jnp.sin(10 * jnp.pi * x[..., 0]) / 2 * x[..., 0] + (x[..., 0] - 1) ** 4


def hartmann_6d(x: Array) -> Numeric:
  alpha = jnp.array([1.0, 1.2, 3.0, 3.2])

  A = jnp.array(
    [
      [10, 3, 17, 3.5, 1.7, 8],
      [0.05, 10, 17, 0.1, 8, 14],
      [3, 3.5, 1.7, 10, 17, 8],
      [17, 8, 0.05, 10, 0.1, 14],
    ]
  )

  P = jnp.array(
    [
      [1312, 1696, 5569, 124, 8283, 5886],
      [2329, 4135, 8307, 3736, 1004, 9991],
      [2348, 1451, 3522, 2883, 3047, 6650],
      [4047, 8828, 8732, 5743, 1091, 381.0],
    ]
  )

  inner_sum = jnp.sum(
    A * (x[jnp.newaxis, jnp.newaxis, ...] - 0.0001 * P) ** 2, axis=-1
  )

  return -jnp.sum(alpha * jnp.exp(-inner_sum), axis=-1)



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for parameters."""

from typing import NamedTuple

from boax.experiments.search_spaces.parameters.alias import (
  Choice,
  LogRange,
  Range,
)


class SearchSpace(NamedTuple):
  """
  A tuple describing an experiment search space.

  Attributes:
    choice_parameters: A list of all choice parameters in the search space.
    range_parameters: A list of all range and log_range parameters in the search space.
  """

  choice_parameters: list[Choice]
  range_parameters: list[LogRange | Range]



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for sampling functions."""

import math

from jax import numpy as jnp
from jax import random

from boax.core import distributions
from boax.core.distributions.multivariate_normal import MultivariateNormal
from boax.core.distributions.normal import Normal
from boax.core.distributions.uniform import Uniform
from boax.core.samplers import functions
from boax.core.samplers.base import Sampler
from boax.utils.typing import Array, PRNGKey, Shape


def uniform(key: PRNGKey, sample_shape: Shape) -> Sampler[Uniform]:
  """
  The i.i.d. uniform sampler.

  Example:
    >>> sampler = uniform(key, (128,))
    >>> samples = sampler(Uniform(0, 1))

  Args:
    key: The pseudo-random number genererator key.
    sample_shape: The sample shape.

  Returns:
    The corresponding `Sampler`.
  """

  def sample(uniform: Uniform) -> Array:
    batch_shape = uniform.a.shape[:-1]
    event_shape = jnp.broadcast_shapes(uniform.a.shape[-1], uniform.b.shape[-1])
    base_samples = random.uniform(key, sample_shape + batch_shape + event_shape)

    return distributions.uniform.sample(uniform, base_samples)

  return sample


def normal(key: PRNGKey, sample_shape: Shape) -> Sampler[Normal]:
  """
  The i.i.d. normal sampler.

  Example:
    >>> sampler = normal(key, (128,))
    >>> samples = sampler(Normal(0, 1))

  Args:
    key: The pseudo-random number genererator key.
    sample_shape: The sample shape.

  Returns:
    The corresponding `Sampler`.
  """

  def sample(normal: Normal) -> Array:
    batch_shape = normal.loc.shape[:-1]
    event_shape = jnp.broadcast_shapes(
      normal.loc.shape[-1], normal.scale.shape[-1]
    )
    base_samples = random.normal(key, sample_shape + batch_shape + event_shape)

    return distributions.normal.sample(normal, base_samples)

  return sample


def multivariate_normal(
  key: PRNGKey, sample_shape: Shape
) -> Sampler[MultivariateNormal]:
  """
  The i.i.d. multivariate normal sampler.

  Example:
    >>> sampler = multivariate_normal(key, (128,))
    >>> samples = sampler(MultivariateNormal(jnp.zeros((1,)), jnp.identity(1)))

  Args:
    key: The pseudo-random number genererator key.
    sample_shape: The sample shape.

  Returns:
    The corresponding `Sampler`.
  """

  def sample(mvn: MultivariateNormal) -> Array:
    batch_shape = mvn.mean.shape[:-1]
    event_shape = jnp.broadcast_shapes(mvn.mean.shape[-1], mvn.cov.shape[-2])
    base_samples = random.normal(key, sample_shape + batch_shape + event_shape)

    return distributions.multivariate_normal.sample(mvn, base_samples)

  return sample


def halton_uniform(key: PRNGKey, sample_shape: Shape) -> Sampler[Uniform]:
  """
  The quasi-MC uniform sampler based on halton sequences.

  Example:
    >>> sampler = halton_uniform(key, (128,))
    >>> samples = sampler(Uniform(0, 1))

  Args:
    key: The pseudo-random number genererator key.
    sample_shape: The sample shape.

  Returns:
    The corresponding `Sampler`.
  """

  def sample(uniform: Uniform) -> Array:
    batch_shape = uniform.a.shape[:-1]
    event_shape = jnp.broadcast_shapes(uniform.a.shape[-1], uniform.b.shape[-1])
    out_shape = batch_shape + event_shape
    ndims = math.prod(out_shape)

    if ndims < 1 or ndims > functions.quasi_random.MAX_DIMENSION:
      raise ValueError(
        'Dimension must be between 1 and {}. Supplied {}'.format(
          functions.quasi_random.MAX_DIMENSION, ndims
        )
      )

    base_samples = functions.quasi_random.halton_sequence(
      key, sample_shape, ndims
    )
    reshaped = jnp.reshape(base_samples, sample_shape + out_shape)

    return distributions.uniform.sample(uniform, reshaped)

  return sample


def halton_normal(key: PRNGKey, sample_shape: Shape) -> Sampler[Normal]:
  """
  The quasi-MC normal sampler based on halton sequences.

  Example:
    >>> sampler = halton_normal(key, (128,))
    >>> samples = sampler(Normal(0, 1))

  Args:
    key: The pseudo-random number genererator key.
    sample_shape: The sample shape.

  Returns:
    The corresponding `Sampler`.
  """

  def sample(normal: Normal) -> Array:
    out_shape = jnp.broadcast_shapes(normal.loc.shape, normal.scale.shape)
    ndims = math.prod(out_shape)

    if ndims < 1 or ndims > functions.quasi_random.MAX_DIMENSION:
      raise ValueError(
        'Dimension must be between 1 and {}. Supplied {}'.format(
          functions.quasi_random.MAX_DIMENSION, ndims
        )
      )

    base_samples = functions.quasi_random.halton_sequence(
      key, sample_shape, ndims
    )
    normal_samples = functions.utils.ratio_of_uniforms(base_samples)
    reshaped = jnp.reshape(normal_samples, sample_shape + out_shape)

    return distributions.normal.sample(normal, reshaped)

  return sample


def halton_multivariate_normal(
  key: PRNGKey, sample_shape: Shape
) -> Sampler[MultivariateNormal]:
  """
  The quasi-MC multivariate normal sampler based on halton sequences.

  Example:
    >>> sampler = halton_multivariate_normal(key, (128,))
    >>> samples = sampler(MultivariateNormal(jnp.zeros((1,)), jnp.identity(1)))

  Args:
    key: The pseudo-random number genererator key.
    sample_shape: The sample shape.

  Returns:
    The corresponding `Sampler`.
  """

  def sample(mvn: MultivariateNormal) -> Array:
    batch_shape = mvn.mean.shape[:-1]
    event_shape = jnp.broadcast_shapes(mvn.mean.shape[-1], mvn.cov.shape[-2])
    out_shape = batch_shape + event_shape
    ndims = math.prod(out_shape)

    if ndims < 1 or ndims > functions.quasi_random.MAX_DIMENSION:
      raise ValueError(
        'Dimension must be between 1 and {}. Supplied {}'.format(
          functions.quasi_random.MAX_DIMENSION, ndims
        )
      )

    base_samples = functions.quasi_random.halton_sequence(
      key, sample_shape, ndims
    )
    normal_samples = functions.utils.ratio_of_uniforms(base_samples)
    reshaped = jnp.reshape(normal_samples, sample_shape + out_shape)

    return distributions.multivariate_normal.sample(mvn, reshaped)

  return sample



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for experiments."""

from typing import Any, Generic, NamedTuple, Protocol, TypeVar

T = TypeVar('T')


class Step(Generic[T], NamedTuple):
  """
  A tuple describing a single step of an experiment trial.

  Attributes:
    timestep: The step's timestep.
    state: The step's current state.
  """

  timestep: int
  state: T


class NextFn(Generic[T], Protocol):
  """
  A callable for running a single optimization step.
  """

  def __call__(
    self,
    step: Step[T] | None = None,
    results: list[tuple[dict[str, Any], float]] = [],
  ) -> tuple[Step[T], list[dict[str, Any]]]:
    """
    Returns the next step and a list of paramterizations to explore.

    Args:
      step: The current step object.
      results: Results from evaluating the current step parameterizations.

    Returns:
      A tuple of the next step and the paramterizations to explore at this step.
    """


class BestFn(Generic[T], Protocol):
  """
  A callable predicting the currenlty best parameterization.
  """

  def __call__(self, step: Step[T]) -> tuple[dict[str, Any], float]:
    """
    Returns the best predicted parameterization and its predicted value.

    Args:
      step: The current step object.

    Returns:
      A tuple of the best predicted parameterization and its predicted value.
    """


class Trial(Generic[T], NamedTuple):
  """
  A tuple describing an experiment trial.

  Attributes:
    next: A callable running a single optimization step.
    best: A callable predicting the currenlty best parameterization.
  """

  next: NextFn[T]
  best: BestFn[T]



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for policy functions."""

from typing import Generic, Protocol, TypeVar

from boax.utils.typing import PRNGKey

T = TypeVar('T')


class Policy(Protocol, Generic[T]):
  """
  A callable type for policy functions.

  A policy function takes a set of parameters of type `T`,
  a timestep, and a pseudo-random key as input and returns a selected variant.
  """

  def __call__(self, params: T, timestep: int, key: PRNGKey) -> int:
    """
    Select the variant.

    Args:
      params: The policy's parameters.
      timestep: The current timestep.
      key: A pseudo-random key.

    Returns:
      The selected variant.
    """



# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Alias for likelihoods."""

from functools import partial

from jax import jit

from boax.core.distributions.beta import Beta
from boax.core.distributions.multivariate_normal import MultivariateNormal
from boax.core.models.likelihoods import functions
from boax.core.models.likelihoods.base import Likelihood
from boax.utils.typing import Array, Numeric


def gaussian(
  noise: Numeric,
) -> Likelihood[MultivariateNormal, MultivariateNormal]:
  """
  The gaussian likelihood function.

  Example:
    >>> likelihood = gaussian(1e-4)
    >>> predictive = likelihood(mvn)

  Args:
    noise: The noise parameter.

  Returns:
    The gaussian `Likelihood` function.
  """

  return jit(
    partial(
      functions.marginal.gaussian,
      noise=noise,
    ),
  )


def beta(scale: Numeric) -> Likelihood[Array, Beta]:
  """
  The beta likelihood function.

  Example:
    >>> likelihood = beta(1e-4)
    >>> predictive = likelihood(xs)

  Args:
    scale: The scale parameter.

  Returns:
    The beta `Likelihood` function.
  """

  return jit(
    partial(
      functions.conditional.beta,
      scale=scale,
    ),
  )
